unit uFileList;
{$INCLUDE 'FRename3.inc'}

interface

uses
  Windows, SysUtils, Classes, Generics.Collections, Generics.Defaults,
  {$INCLUDE 'DelphiFarUnits.inc'} ,
  uOptions,
  uLog, uSystem, uFiles, uMessages, uFRenameConsts;

type
  TFileList = class(TFarEditorManaged)
  protected
    type TKeyEventType = (keNone, keAltF1, keAltF2);
  private
    fSourceDirectory: string;
    fOptions: TOptions;
    fRepeatedRun: boolean;
  protected
    procedure AddFileName(Text: TTextStream; const FileName: string; MaxFileNameLength: integer);
    function GetTranslatedFileName(const FileName: string): string;
    function DecodeEditorLine(const Line: string; out SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer): boolean; overload;
    function DecodeEditorLine(const Line: string; out SourceFileName, DestinationFileName: string): boolean; overload;
    function GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
    procedure BeforeClose(var CanClose: boolean; var Result: TIntPtr); override;
    procedure AfterReOpen; override;
    class function GetKeyEventType(const Info: TKeyEventRecord): TKeyEventType;
    class function CurrentEditorKeyEvent(const Info: TKeyEventRecord): boolean; override;
    function ProcessKeyEvent(const Info: TKeyEventRecord): boolean; override;
    function GetDescription: string;
    property RepeatedRun: boolean read fRepeatedRun;
  public
    class function FindByEditorID(EditorID: TIntPtr; out Editor: TFileList): boolean;
    constructor Create(Options: TOptions);
    destructor Destroy; override;
    function LoadFromPanel(AActivePanel: boolean): boolean; overload;
    function LoadFromPanel(const AInfo: TFarPanel): boolean; overload;
    function LoadFromFileList(const Directory: string; const Files: TStrings): boolean;
    function OpenInEditor: boolean;
    function RenameFiles: boolean;
    function SelectDestinationFilenames: boolean;
    function JumpToDestinationFileName: boolean;
    function ShowEditorMenu: boolean;
    property SourceDirectory: string read fSourceDirectory write fSourceDirectory;
    property Options: TOptions read fOptions;
  end;

implementation

type
  TActiveFileLists = class(TObjectList<TFileList>);

var
  ActiveFileLists: TActiveFileLists = nil;

{ TFileList }

procedure TFileList.AddFileName(Text: TTextStream; const FileName: string; MaxFileNameLength: integer);
begin
  Text.WriteLine(Format('%-*.*s    %s', [MaxFileNameLength, MaxFileNameLength, FileName, FileName]));
end;

procedure TFileList.AfterReOpen;
begin
  inherited;
  fRepeatedRun := True;
end;

procedure TFileList.BeforeClose(var CanClose: boolean; var Result: TIntPtr);
begin
  inherited;
  try
    CanClose := RenameFiles;
  except
    on EAbort do
      CanClose := True;
  end;
end;

constructor TFileList.Create(Options: TOptions);
begin
  inherited Create(SystemTempFile(SystemTempDir, 'FRename3', '.txt'));
  ActiveFileLists.Add(Self);
  fOptions := Options;
end;

class function TFileList.CurrentEditorKeyEvent(const Info: TKeyEventRecord): boolean;
begin
  Result := False;
  if GetKeyEventType(Info) <> keNone then
    Result := inherited CurrentEditorKeyEvent(Info);
end;

function TFileList.DecodeEditorLine(const Line: string; out SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer): boolean;
var
  i, n: integer;
begin
  Result := False;
  n := Length(Line);
  if n > 0 then
    if Line[1] = '"' then begin
      SourceFileNameStart := 2;
      i := 2;
      while (i <= n) and (Line[i] <> '"') do
        Inc(i);
      SourceFileNameLength := i - SourceFileNameStart;
      Inc(i);
      while (i <= n) and (Line[i] = ' ') do
        Inc(i);
      if (i <= n) and (Line[i] = '"') then begin
        Inc(i);
        DestinationFileNameStart := i;
        while (i <= n) and (Line[i] <> '"') do
          Inc(i);
        DestinationFileNameLength := i - DestinationFileNameStart;
        Inc(i);
        while (i <= n) and (Line[i] = ' ') do
          Inc(i);
        if Pred(i) = n then
          Result := True;
        end;
      end;
end;

function TFileList.DecodeEditorLine(const Line: string; out SourceFileName, DestinationFileName: string): boolean;
var
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
begin
  Result := False;
  if DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength) then
    if SourceFileNameLength > 0 then
      if DestinationFileNameLength > 0 then begin
        SourceFileName := Copy(Line, SourceFileNameStart, SourceFileNameLength);
        DestinationFileName := Copy(Line, DestinationFileNameStart, DestinationFileNameLength);
        Result := True;
      end;
end;

destructor TFileList.Destroy;
var
  i: Integer;
begin
  for i := Pred(ActiveFileLists.Count) downto 0 do
    if ActiveFileLists[i] = Self then
      ActiveFileLists.Delete(i);
  if FileName <> '' then
    DeleteFile(FileName);
  inherited;
end;

class function TFileList.FindByEditorID(EditorID: TIntPtr; out Editor: TFileList): boolean;
begin
  Result := inherited FindByEditorID(EditorID, TFarEditorManaged(Editor));
end;

function TFileList.GetDescription: string;
begin
  if Options.TinyHelp then
    Result := GetMsg(MFileListDescriptionShort)
  else
    Result := GetMsg(MFileListDescription);
end;

function TFileList.GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
var
  LastError: DWORD;
  SR: TSearchRec;
begin
  Attributes := Windows.GetFileAttributes(PChar(FileName));
  if Attributes <> INVALID_FILE_ATTRIBUTES then
    Result := True
  else begin
    LastError := GetLastError;
    if (LastError = ERROR_FILE_NOT_FOUND) or (LastError = ERROR_PATH_NOT_FOUND) or (LastError = ERROR_INVALID_NAME) then
      Result := False
    else
      if FindFirst(FileName, faAnyFile, SR) = 0 then
        try
          Attributes := SR.Attr;
          Result := True;
        finally
          SysUtils.FindClose(SR);
        end
      else
        Result := False;
  end;
end;

class function TFileList.GetKeyEventType(const Info: TKeyEventRecord): TKeyEventType;
var
  ShiftFlags: DWORD;
begin
  Result := keNone;
  if Info.bKeyDown then
  begin
    ShiftFlags := Info.dwControlKeyState and (LEFT_ALT_PRESSED or LEFT_CTRL_PRESSED or RIGHT_ALT_PRESSED or RIGHT_CTRL_PRESSED or SHIFT_PRESSED);
    if (ShiftFlags = LEFT_ALT_PRESSED) or (ShiftFlags = RIGHT_ALT_PRESSED) then
    begin
      if Info.wVirtualKeyCode = VK_F1 then
        Result := keAltF1
      else if Info.wVirtualKeyCode = VK_F2 then
        Result := keAltF2
      ;
    end;
  end;
end;

function TFileList.GetTranslatedFileName(const FileName: string): string;
begin
  Result := AnsiQuotedStr(FileName, '"');
end;

function TFileList.JumpToDestinationFileName: boolean;
var
  EditorInfo: TEditorInfo;
  Line: string;
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
begin
  Result := False;
  if TFarEditor.GetInfo(EditorID, EditorInfo) then
    if TFarEditor.GetString(EditorID, EditorInfo.CurLine, Line) then
      if DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength) then
        if TFarEditor.SetCursorPosition(EditorID, EditorInfo.CurLine, DestinationFileNameStart-1) then begin
          TFarEditor.Refresh(EditorID);
          Result := True;
        end;
end;

function TFileList.LoadFromPanel(AActivePanel: boolean): boolean;
var
  Panel: TFarPanel;
begin
  Panel := TFarPanel.Create(AActivePanel);
  try
    Result := LoadFromPanel(Panel);
  finally
    FreeAndNil(Panel);
  end;
end;

function TFileList.LoadFromFileList(const Directory: string; const Files: TStrings): boolean;
var
  Text: TTextStream;
  Description: string;
  i, n, MaxLength: integer;
begin
  Result := False;
  MaxLength := 0;
  for i := 0 to Pred(Files.Count) do
  begin
    n := Length(GetTranslatedFileName(Files[i]));
    if n > MaxLength then
      MaxLength := n;
  end;
  if MaxLength > 0 then begin
    Description := GetDescription;
    Description := StringReplace(Description, '${DIRECTORY}', Directory, [rfReplaceAll]);
    SourceDirectory := IncludeTrailingPathDelimiter(FullFileName(Directory, True));
    Text := TTextStream.Create(FileName, fmCreate);
    try
      Text.Encoding := feUtf8;
      if Description <> '' then
        Text.WriteLine(Description);
      for i := 0 to Pred(Files.Count) do
      begin
        AddFileName(Text, GetTranslatedFileName(Files[i]), MaxLength);
      end;
      Result := True;
    finally
      FreeAndNil(Text);
    end;
  end;
end;

function TFileList.LoadFromPanel(const AInfo: TFarPanel): boolean;
var
  Text: TTextStream;
  Description: string;
  i, n, MaxLength: integer;
  SelectedItems: boolean;
begin
  Result := False;
  MaxLength := 0;
  for i := 0 to Pred(AInfo.ItemCount) do
    if AInfo.Items[i].Selected then begin
      n := Length(GetTranslatedFileName(AInfo.Items[i].FileName));
      if n > MaxLength then
        MaxLength := n;
    end;
  SelectedItems := MaxLength > 0;
  if not SelectedItems then begin
    MaxLength := Length(GetTranslatedFileName(AInfo.Items[AInfo.CurrentItemIndex].FileName));
  end;
  if MaxLength > 0 then begin
    Description := GetDescription;
    Description := StringReplace(Description, '${DIRECTORY}', AInfo.Directory, [rfReplaceAll]);
    SourceDirectory := IncludeTrailingPathDelimiter(FullFileName(AInfo.Directory, True));
    Text := TTextStream.Create(FileName, fmCreate);
    try
      Text.Encoding := feUtf8;
      if Description <> '' then
        Text.WriteLine(Description);
      if SelectedItems then begin
        for i := 0 to Pred(AInfo.ItemCount) do begin
          if AInfo.Items[i].Selected then begin
            AddFileName(Text, GetTranslatedFileName(AInfo.Items[i].FileName), MaxLength);
          end;
        end;
      end
      else begin
        AddFileName(Text, GetTranslatedFileName(AInfo.Items[AInfo.CurrentItemIndex].FileName), MaxLength);
      end;
      Result := True;
    finally
      FreeAndNil(Text);
    end;
  end;
end;

function TFileList.OpenInEditor: boolean;
{$IFDEF FAR3_UP}
const
  KEY_ALT_F1: TFarKey = (VirtualKeyCode: VK_F1; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F2: TFarKey = (VirtualKeyCode: VK_F2; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
{$ENDIF}
var
  Description: string;
  i, DescriptionLines: integer;
  Labels: TFarEditor.TEditorKeyBarLabels;
begin
  Result := False;
  if (FileName <> '') and FileExists(FileName) then
  begin
    Description := GetDescription;
    if Description = '' then
    begin
      DescriptionLines := 0;
    end
    else
    begin
      DescriptionLines := 1;
      for i := 1 to Length(Description) do
        if Description[i] = #10 then
          Inc(DescriptionLines);
    end;
    if Show(GetMsg(MTitle), 0, 0, -1, -1,
          EF_NONMODAL or EF_IMMEDIATERETURN or EF_DELETEONLYFILEONCLOSE or EF_DISABLEHISTORY {$IFDEF FAR3_UP} or EF_DISABLESAVEPOS {$ENDIF} ,
          DescriptionLines+1, 0, CP_UTF8)
    then
    begin
      JumpToDestinationFileName;
      SetLength(Labels, 2);
      {$IFDEF FAR3_UP}
      Labels[0].Key := KEY_ALT_F1;
      {$ENDIF}
      Labels[0].Caption := GetMsg(MKeyJumpToDestinationFileName);
      Labels[0].Description := GetMsg(MKeyJumpToDestinationFileNameLong);
      {$IFDEF FAR3_UP}
      Labels[1].Key := KEY_ALT_F2;
      {$ENDIF}
      Labels[1].Caption := GetMsg(MKeySelectDestinationFileNames);
      Labels[1].Description := GetMsg(MKeySelectDestinationFileNamesLong);
      SetKeyBar(Labels);
      Result := True;
    end
    else
      ShowWarning(GetMsg(MFailedToOpenEditor));
  end;
end;

function TFileList.ProcessKeyEvent(const Info: TKeyEventRecord): boolean;
begin
  Result := False;
  case GetKeyEventType(Info) of
    keAltF1:
      Result := JumpToDestinationFileName;
    keAltF2:
      Result := SelectDestinationFilenames;
  end;
end;

function TFileList.RenameFiles: boolean;
var
  Text: TTextStream;
  FailedText: TTextStream;
  FailedStream: TMemoryStream;
  Line, SourceFileName, DestinationFileName, SrcFN, DestFN: string;
  SrcAttr, DestAttr: DWORD;
  Again, AlwaysSkip: boolean;
  Failed: boolean;
  Panel: TFarPanel;
  i, FileIndex: Integer;
begin
  Result := False;
  if (FileName <> '') and FileExists(FileName) then begin
    Panel := nil;
    FailedText := nil;
    Text := TTextStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    try
      FailedStream := TMemoryStream.Create;
      FailedText := TTextStream.Create(FailedStream, True);
      AlwaysSkip := False;
      Result := True;
      if Options.SelectRenamedFiles then
      begin
        Panel := TFarPanel.Create(True, False);
        if AnsiCompareText(IncludeTrailingPathDelimiter(FullFileName(Panel.Directory, True)), SourceDirectory) <> 0 then
          FreeAndNil(Panel);
        if Panel <> nil then
        begin
          Panel.Refresh;
          if not RepeatedRun then
          begin
            for i := 0 to Pred(Panel.ItemCount) do
              Panel.Items[i].Selected := False;
            Panel.ApplySelection;
          end;
        end;
      end;
      while Text.ReadLine(Line) do
      begin
        Failed := True;
        if DecodeEditorLine(Line, SourceFileName, DestinationFileName) then
        begin
          repeat
            Again := False;
            SrcFN := SourceDirectory + SourceFileName;
            DestFN := SourceDirectory + DestinationFileName;
            if SrcFN = DestFN then
            begin
              Failed := False;
              Break;
            end
            else if not GetFileAttributes(SrcFN, SrcAttr) then
              ShowMessage([GetMsg(MTitle), GetMsg(MSourceFileNotFound), SrcFN, GetMsg(MOk)], FMSG_WARNING, 1)
            else begin
              if GetFileAttributes(DestFN, DestAttr) and ((DestAttr and FILE_ATTRIBUTE_DIRECTORY) <> 0) and (AnsiCompareText(SrcFN, DestFN) <> 0) then begin
                ShowMessage([GetMsg(MTitle), GetMsg(MCantOverwriteDirectory), DestFN, GetMsg(MOk)], FMSG_WARNING, 1);
                Result := False;
              end
              else if not RenameFile(SrcFN, DestFN) then
              begin
                if AlwaysSkip then
                  Result := False
                else
                begin
                  case ShowMessage([GetMsg(MTitle), GetMsg(MFailedToRename), SrcFN, GetMsg(MFailedToRenameTo), DestFN, GetMsg(MRetry), GetMsg(MOverwrite), GetMsg(MSkip), GetMsg(MSkipAll)], FMSG_WARNING, 4) of
                    0:   Again := True;
                    1:   begin
                           DeleteFile(DestFN);
                           Again := True;
                         end;
                    2:   Result := False;
                    3:   begin
                           AlwaysSkip := True;
                           Result := False;
                         end;
                    else Abort;
                  end;
                end
              end
              else
                Failed := False;
            end;
          until not Again;
        end;
        if Failed then
        begin
          FailedText.WriteLine(Line);
        end
        else if Panel <> nil then
        begin
          Panel.Refresh;
          if Panel.FindFile(DestinationFileName, FileIndex) then
          begin
            Panel.Items[FileIndex].Selected := True;
            Panel.ApplySelection;
          end;
        end;
      end;
      if not Result then
      begin
        FreeAndNil(Text);
        FailedText.Flush;
        FailedStream.SaveToFile(FileName);
      end;
    finally
      FreeAndNil(Text);
      FreeAndNil(FailedText);
      FreeAndNil(Panel);
      TFarPanel.Refresh(True);
    end;
  end;
end;

function TFileList.SelectDestinationFilenames: boolean;
var
  EditorInfo: TEditorInfo;
  Select: TEditorSelect;
  LineNr, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
  Line: string;
begin
  Result := False;
  if TFarEditor.GetInfo(EditorID, EditorInfo) then begin
    FillChar(Select, Sizeof(Select), 0);
    {$IFDEF FAR3_UP}
    Select.StructSize := Sizeof(Select);
    {$ENDIF}
    Select.BlockType := BTYPE_COLUMN;
    Select.BlockStartLine := -1;
    Select.BlockStartPos := -1;
    Select.BlockWidth := -1;
    Select.BlockHeight := -1;
    for LineNr := 0 to Pred(EditorInfo.TotalLines) do
      if TFarEditor.GetString(EditorID, LineNr, Line) then
        if DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength) then begin
          if Select.BlockStartLine < 0 then
            Select.BlockStartLine := LineNr;
          if (Select.BlockStartPos < 0) or (Select.BlockStartPos > (DestinationFileNameStart-2)) then
            Select.BlockStartPos := DestinationFileNameStart-2;
          if Select.BlockWidth < (DestinationFileNameLength+2) then
            Select.BlockWidth := DestinationFileNameLength+2;
          Select.BlockHeight := LineNr - Select.BlockStartLine + 1;
          Result := True;
        end;
    if Result then begin
      Result := FarApi.EditorControl( {$IFDEF FAR3_UP} EditorID, {$ENDIF} ECTL_SELECT, {$IFDEF FAR3_UP} 0, {$ENDIF} @Select) <> 0;
      TFarEditor.Refresh(EditorID);
    end;
  end;
end;

function TFileList.ShowEditorMenu: boolean;
var
  SelectedItem: integer;
begin
  Result := False;
  if ShowMenu(MENU_EDITOR_GUID, GetMsg(MTitle), '', [GetMsg(MKeyJumpToDestinationFileNameLong), GetMsg(MKeySelectDestinationFileNamesLong)], SelectedItem) then
  begin
    case SelectedItem of
      0:   Result := JumpToDestinationFileName;
      1:   Result := SelectDestinationFilenames;
    end;
  end;
end;

initialization
  ActiveFileLists := TActiveFileLists.Create(False);

finalization
  FreeAndNil(ActiveFileLists);

end.

